# -*- coding: utf-8 -*

from flyai.core import Lib


class Dataset:
    def __init__(self, epochs=5, batch=32, val_batch=32):
        """
        :param epochs: 训练的轮次，最大不超过100
        :param batch: 训练的批次大小，太大会导致显存不足
        :param val_batch: 验证的批次大小
        """
        self.lib = Lib(epochs, batch, val_batch)

    def image_aug(self,
                  target_size,
                  epochs=1,
                  seed=None,
                  featurewise_center=False,
                  samplewise_center=False,
                  featurewise_std_normalization=False,
                  samplewise_std_normalization=False,
                  zca_whitening=False,
                  zca_epsilon=1e-6,
                  rotation_range=0,
                  width_shift_range=0.,
                  height_shift_range=0.,
                  brightness_range=None,
                  shear_range=0.,
                  zoom_range=0.,
                  channel_shift_range=0.,
                  fill_mode='nearest',
                  cval=0.,
                  horizontal_flip=False,
                  vertical_flip=False,
                  rescale=None,
                  preprocessing_function=None,
                  aug_label=False):
        """
        为图片做增强的数据，加入到训练的csv中，使用dataset.next_train_batch()可读取到增强后的数据
        :param target_size: 增强图片的大小 如(256,256)
        :param epochs: 对图片增强几轮，每轮一张图片。
        :param batch_size: 每次增强多少张图片。
        :param seed: 进行随机图像变换, 通过设置seed可以达到同步变换。
        :param featurewise_center: 布尔值，对输入的图片每个通道减去每个通道对应均值。
        :param samplewise_center: 布尔值，每张图片减去样本均值, 使得每个样本均值为0。
        :param featurewise_std_normalization: 布尔值，将输入除以数据集的标准差以完成标准化
        :param samplewise_std_normalization: 布尔值，将输入的每个样本除以其自身的标准差
        :param zca_whitening: 布尔值，对输入数据施加ZCA白化
        :param zca_epsilon:ZCA白化值。默认值为1e-6。
        :param rotation_range: 整数，数据提升时图片随机转动的角度
        :param width_shift_range: 浮点数，图片宽度的某个比例，数据提升时图片水平偏移的幅度
        :param height_shift_range: 浮点数，图片高度的某个比例，数据提升时图片竖直偏移的幅度
        :param brightness_range: Tuple 或者 list对数据进行亮度处理
        :param shear_range: 浮点数，剪切强度（逆时针方向的剪切变换角度）
        :param zoom_range: 浮点数或形如[lower,upper]的列表，随机缩放的幅度，若为浮点数，则相当于[lower,upper] = [1 - zoom_range, 1+zoom_range]
        :param channel_shift_range: 浮点数，随机通道偏移的幅度
        :param fill_mode: constant，nearest，reflect或wrap之一，当进行变换时超出边界的点将根据本参数给定的方法进行处理
        :param cval: 浮点数或整数，当fill_mode=constant时，指定要向超出边界的点填充的值
        :param horizontal_flip: 布尔值，进行随机水平翻转
        :param vertical_flip: 布尔值，进行随机竖直翻转
        :param rescale: 重放缩因子,默认为None. 如果为None或0则不进行放缩,否则会将该数值乘到数据上(在应用其他变换之前)
        :param preprocessing_function: 增强后自定义处理函数
        :param aug_label: 是否对label标签增强，一般应用在图像分割等领域。
        :return:
        """
        self.lib.image_aug(
            target_size=target_size,
            epochs=epochs,
            seed=seed,
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening,
            zca_epsilon=zca_epsilon,
            rotation_range=rotation_range,
            width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range,
            shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range,
            fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip,
            vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            aug_label=aug_label)

    def add_aug_data(self, data):
        """
        把自自己增强的数据，加入到训练的csv中，使用dataset.next_train_batch()可读取到增强后的数据
        :param data: data的数据格式为：
                    单条数据
                    {
                    "image_path": ["images/skunk/cifar100_48287_75.png"],

                    "label":[75]
                    }
                    多条数据
                    {
                    "image_path": ["images/skunk/cifar100_48287_75.png","images/bicycle/cifar100_38531_8.png"],

                    "label": [75,8]
                    }
        """
        self.lib.add_aug_data(data)

    def get_step(self):
        """
        根据dataset传入的epochs和batch，计算出来的训练总次数。
        :return: 返回训练总次数
        """
        return self.lib.get_step()

    def get_train_length(self):
        """
        获取训练集总数量，本地调用返回的是100条，在GPU上调用返回全部数据集数量。
        :return: 返回训练集总数量
        """
        return self.lib.get_train_length()

    def get_validation_length(self):
        """
        获取验证集总数量，本地调用返回的是100条，在GPU上调用返回全部数据集数量。
        :return: 返回验证集总数量
        """
        return self.lib.get_validation_length()

    def next_train_batch(self):
        """
        获取一批训练数据，返回数据的数量是dataset中batch的大小。
        :return: x_train,y_train
        """
        return self.lib.next_train_batch()

    def next_validation_batch(self):
        """
        获取一批验证数据，返回数据的数量是dataset中val_batch的大小。
        :return: x_val,y_val
        """
        return self.lib.next_validation_batch()

    def next_batch(self, size=32, test_size=32, test_data=True):
        """
        获取一批训练和验证数据，可以自己设置返回的大小。
        :return:x_train,y_train,x_val,y_val
        """
        return self.lib.next_batch(size, test_size, test_data)

    # def create_generator(self, x_data, y_data, batch=32):
    #     """
    #     创建生成器迭代数据
    #     :param x_data: 训练的X数据
    #     :param y_data: 训练的Y数据
    #     :param batch: 每次生成数据的大小
    #     :return:返回训练数据的生成器
    #     """
    #     return self.lib.create_generator(x_data, y_data, batch)

    def get_all_processor_data(self):
        """
        获取所有在processor.py中，通过input_x方法处理过的数据
        :return:x_train,y_train,x_val,y_val
        """
        return self.lib.get_all_processor_data()

    def get_all_data(self):
        """
        获取所有原始数据
        :return:x_train,y_train,x_val,y_val
        """
        return self.lib.get_all_data()

    def get_all_validation_data(self):
        """
        获取所有在processor.py中，通过input_x方法处理过的验证集数据
        :return:x_val,y_val
        """
        return self.lib.get_all_validation_data()

    def evaluate_data_no_processor(self, path=None):
        """
        获取所有原始验证集数据
        :return:x_val,y_val
        """
        return self.lib.evaluate_data_no_processor(path)

    def output_y(self, predict):
        return self.lib.to_categorys(predict)

    def evaluate_source(self, path):
        return self.lib.evaluate_source(path)

    def evaluate_data(self, path=None):
        return self.lib.evaluate_data(path)

    def processor_x(self, x_datas):
        return self.lib.processor_x(x_datas)

    def processor_y(self, y_datas):
        return self.lib.processor_y(y_datas)

    def predict_data(self, **data):
        return self.lib.predict_data(**data)

    def to_categorys(self, predict):
        return self.lib.to_categorys(predict)

    def create_instance(self, module_name, class_name, *args, **kwargs):
        return self.lib.create_instance(module_name, class_name, *args, **kwargs)

    def get_method_dict(self, clz, method_name, **args):
        return self.lib.get_method_dict(clz, method_name, **args)

    def get_method_list(self, clz, method_name, *args):
        return self.lib.get_method_list(clz, method_name, *args)
